import java.util.*;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;


/* A base class for ITEC120 (CS1) that provides two things:
 * - Before introducing objects: a set of utility methods
 *   (mostly, static versions of String methods)
 * - When first introducing object, extending Object120 will auto-provide
 *   a constructor, equals, toString (and, hashcode).
 *   The constructor takes in one argument per field;
 *   the toString returns a String which looks the same as a constructor call;
 *   the equals does a deep comparison of each fields.
 * @author Ian Barland
 * @version 2010.Mar.03
 */
abstract class Object120 {

  /** Return the version number of this library
   * (since it might get updated through the semester).
   */
  public static String getVersion() { return "class Object120: v1.0 (Aardvark), 2010-Aug-28."; }
  public static void printVersion() { System.err.println( Object120.getVersion() ); }
  //static { Object120.printVersion(); }
 
  /** Constructor:
   * Initialize each field with the provided args,
   * in the declared order.
   */
  public Object120( Object[] args ) {
    // TODO: for efficiency, keep a static Map<Class<? extends Object120>,List<Field>> instanceFields.
    instanceFields = new ArrayList<Field>();
    for (Field f : this.getClass().getDeclaredFields()) {
      /* N.B. The docs for getDeclaredFields don't explicitly guarantee that two calls
       * on the same Class will return the fields in the same order.  I can't imagine
       * that an implementation *wouldn't* do that, but perhaps sort it just in case?
       */
      if (!Modifier.isStatic(f.getModifiers())) { instanceFields.add(f); }
      }

      
    // We could declare Object... args, but we'll stress how our method is abstract:
    // The subclass must have a varargs constructor which calls 'super(args)'.
    //
    String subclassName = this.getClass().getName();  // For use in diagnostic messages.
    if (!Character.isUpperCase( subclassName.charAt(0) )) {
      System.err.print( "Warning:"
                      + " Class name `" + subclassName + "`"
                      + " should start with upper case, by convention."
                      + " --  Object120 super." );
      }
    if (args.length != this.instanceFields.size()) {
      String msg = "Must provide " + this.instanceFields.size() + " args to constructor";
      msg += " `new " + subclassName + "(";
      boolean needComma = false;
      for (Field f : this.instanceFields) {
        if (Modifier.isStatic(f.getModifiers())) continue;
        msg += (needComma ? ", " : "") + "[" + f.getName() + "]";
        needComma = true;
        }
      msg += ")`.";
      throw new IllegalArgumentException( "Object120 super: " + msg);
      }
      

    //Arrays.sort(fields, compareFieldsAlphByName );
    int i = 0;
    for (Field f : this.instanceFields) {
      if (Modifier.isStatic(f.getModifiers())) continue;
      try {
        f.setAccessible(true);
        f.set(this, args[i] );
        ++i;
        }
      catch(IllegalAccessException e) {
        throw new RuntimeException( "Object120 super: "
                                  + "Security manage doesn't allow accessing fields through reflection. "
                                  + "You must write the " + subclassName + " constructor yourself.\n"
                                  + e.toString() );
        }
      catch(IllegalArgumentException e) {
        throw new RuntimeException( "Object120 super: "
                                  + "Can't initialize field " + subclassName + "." + f.getName()
                                  + " with value " + args[i].toString()
                                  + " (of type " + args[i].getClass().getName() + ").\n"
                                  + e.toString() );
        }
      if (!Character.isLowerCase(f.getName().charAt(0))) {
         System.err.println( "Warning:"
                           + " Field name `" +  f.getName() + "` in class `" + subclassName + "`"
                           + " should start with lower case, by convention."
                           + " --  Object120 super." );
        }
      }
    
    }


    Object120() {
      this( new Object[0] );
      }
    
    
   @Override
   /** A generic hash, summing the hash of each field.
    *  This version ASSUMES THE OBJECT IS IMMUTABLE,
    *  caching the result in the field 'hash'.
    *  For mutable objects, remove the 'if hash==0' guard.
    */
   public int hashCode() {
     if (hash == 0) {
       for (Field f : this.instanceFields) {
         try {
           f.setAccessible(true);
           hash *= Math.pow(2,5)-1;  // shuffle the bits (well?). // TODO: 2,5 as named constants.
           hash += f.get(this).hashCode();
           }
         catch(IllegalAccessException e) {
           throw new RuntimeException( "Rats, I can't access the field through reflection. "
                                     + "Try overriding `hashCode` in " + this.getClass().getName() + "\n"
                                     + e.toString() );
           }
         }
       }
     return hash;
     }
   private int hash = 0;

   @Override
   /** A deep equals check: call equals on each field.
    */
   public boolean equals( Object oth ) {
    if ( this == oth ) {
      return true;
      }
    else if ( oth == null || (this.getClass() != oth.getClass()) ) {
      return false;
      }
    else {
      for (Field f : this.instanceFields) {
        try {
          f.setAccessible(true);
          boolean thisNull = (f.get(this)==null);
          boolean thatNull = (f.get(oth )==null);
          if      ( thisNull &&  thatNull) { /* continue */ }
          else if ( thisNull && !thatNull) return false;
          else if (!thisNull &&  thatNull) return false;
          else if (! f.get(this).equals(f.get(oth))) return false;
          else { /* continue */ }
          }        catch(IllegalAccessException e) {
          throw new RuntimeException( "Rats, I can't access the field through reflection. "
                                    + "Try overriding `equals` in " + this.getClass().getName() + "\n"
                                    + e.toString() );
          }
        }
      return true;
      }
    }
    
  @Override
  public String toString() {
    return this.toString(false);
    }
  
  /** Return a String which looks like a constructor call:
   * @param includeFieldNames If true, include the field names along with their values.
   * @return a String which looks just like a constructor call
   *   when `includeFieldNames` is false,
   *   or similar to one when `includeFieldNames` is true.
   * For example:<pre>
   *   class Foo extends Object120 { int n; String s }
   *   new Foo(7,"hi").toString(false) = "new Foo( 7, \"hi\" )"
   *   new Foo(8,"ho").toString(true)  = "new Foo( n=8, s=\"ho\" )"
   *   </pre>
   * Bug: Depending on the Java compiler, the field names might not be
   * given in the same order they are declared in the class.
   * (This is because java.lang.reflect doesn't provide access to the declared order;
   *  however, many implementations happen to to use that order.)
   */
  public String toString( boolean includeFieldNames ) {
    StringBuilder str = new StringBuilder();
    str.append("new " + this.getClass().getName() + "( ");
    boolean needComma = false;  // Need to insert a comma before the next field?
    for (Field f : this.instanceFields) {
      try {
        if (needComma) str.append( ", " );
        Object theVal = f.get(this);
        if (includeFieldNames) str.append( f.getName().toString() ).append("=");
        if (quoteMarks.get(theVal.getClass()) != null) {
          str.append( quoteMarks.get(theVal.getClass()) );
          }
        if (theVal.getClass().isInstance(Object120.class)) {
          str.append(  Object120.class.cast(theVal).toString(includeFieldNames));
          }
        else {
          str.append( theVal.toString());
          }
        if (quoteMarks.get(theVal.getClass()) != null) {
          str.append( quoteMarks.get(theVal.getClass()) );
          }
        needComma = true;
        }
      catch(IllegalAccessException e) {
        throw new RuntimeException( "Rats, I can't access the field through reflection. "
                                  + "Try overriding `" + this.getClass().getName() + "toString`.\n"
                                  + e.toString() );
        }
      }
    str.append( " )" );
    return str.toString();
    }


  /** When toString'ing fields, put quotes around certain types (Strings, chars, etc). */
  private static Map<Class,String> quoteMarks = new HashMap<Class,String>();
  static {
   quoteMarks.put( Character.class,     "'" );
   quoteMarks.put( String.class,        "\"" );
   quoteMarks.put( StringBuffer.class,  "\"" );
   quoteMarks.put( StringBuilder.class, "\"" );
   }
   
    
  // In case we want to sort the fields alphabetically:
  private static Comparator<Field> compareFieldsAlphByName = new Comparator<Field>() {
    /** @Override */
    public int compare(Field f1, Field f2) { return f1.getName().compareTo( f2.getName() ); }
    /* We could upcase the field name, but that'd be ambiguous if a beginner were to
     * have two field names that differed only by case.
     * Better to make them aware of the difference.
     */
    };
  
  final private List<Field> instanceFields;  // The (non-static) fields in the subclass.
  private int debugOutThreshold = 0;
  // methods assert, debugPrint
  /** Like Java's `assert`, except it's a method.
   */
  static void assert120( boolean cndn, String msg ) {
      if (!cndn) System.err.println( msg );
    }

    
    
    
  /** Does a string start with a vowel?  (Helpful for determining "a(n) ___".)
   * @param s Any string.  Must be non-null.
   * @return whether or not `s` has a first letter which is a,e,i,o,u,y (either case).
   * For example:<pre>
   *   startsWithVowel( "ho" ) = false
   *   startsWithVowel( "oh" ) = true
   *   startsWithVowel( "i" )  = true
   *   startsWithVowel( "" )   = false
   *    </pre>
   */
  private static boolean startsWithVowel( String s ) {
    return (s != null && s.length() > 0 && "aeiouy".contains(s.substring(0,1).toLowerCase()));
    }
    
  private static boolean hasPrefix( String s, String pre ) {
    return indexOf(s,pre) == 0;
    }
  private static String removePrefix( String s, String pre ) {
    assert hasPrefix(s,pre);
    return s.substring(pre.length());
    }
  private static String removePrefixIfPresent( String s, String pre ) {
    return hasPrefix(s,pre) ? removePrefix(s,pre) : s;
    }
    

  /** Check whether a reference is null, and throw an error if it is.
   * @param ref The reference to check for being null.
   * @param methodName The name to put at the start of the (potential) error message.
   * @param type A string describing `ref`s type, for the (potential) error message.
   * @throws NullPointerException if `ref`==null.
   * For example:<pre>
   *   checkForNull( "lala", "myMethod", "String" ) -> no effect
   *   checkForNull( null, "myMethod", "String" ) -> throws an exception with
   *     a message like "Object120.myMethod: expected a String but got null."
   *    </pre>
   */  
  private static void checkForNull( Object ref, String methodName, String type ) {
    if (ref == null) throw new NullPointerException( "Object120." + methodName + ": Expected a" + (startsWithVowel(type) ? "n" : "") + " " + type + " but got null.  (Did you forget to initialize a variable?)" );
    }

  /** Are two values equal?  Works on any values (non-null).
   * A static version of Object.equals(Object),
   * for use in CS1 before we introduce objects.
   * @param _this the first  value to compare.  (Must be non-null.)
   * @param  that the second value to compare.  (Must be non-null.)
   * @return true iff `_this` equals `_that`.
   * For example:<pre>
   *    equals( "hello", "howdy" ) = false
   *    equals( "hello", "hello" ) = true
   *    equals( 23, 22+1 ) = true
   * @see Object#equals(Object)
   */
  public static boolean equals( Object _this, Object that ) {
    checkForNull( _this, "equals", "Object" );
    checkForNull(  that, "equals", "Object" );
    return _this.equals(that);
    }
    
  public static <T,S extends T> int compareTo( Comparable<T> _this, S that ) {
    checkForNull( _this, "compareTo", "Object" );
    String    fullTypeName = _this.getClass().toString();
    String    justTypeName = removePrefix( fullTypeName, "class " );
    String naturalTypeName = removePrefixIfPresent( justTypeName, "java.lang." );
    checkForNull( that,
                  "compareTo",
                  naturalTypeName + " (or something extending " + naturalTypeName + ")" );
    return _this.compareTo(that);
    /* It seems natural to also write a complementary version
     *   static int compareTo( S _this, Comparable<T> that )
     * but then users will commonly get 'error - compareTo is ambiguous, both methods match'.
     */
    }
        
  /** Convert any value into a String.
   * A static version of Object.toString(), for use in CS1 before we introduce objects.
   * @param _this The value to stringify.  Cannot be null.
   * @return a String representation of `_this`.
   * For example:<pre>
   *    toString(43) = "43" 
   *    toString(true) = "true"
   *    </pre>
   * @see Object#toString()
   */
  public static String toString( Object _this ) { return _this.toString(); }

  /** Are two Strings equal (up to case, but including punctuation).
   * A static version of String.equalsIgnoreCase(String), for use in CS1 before we introduce objects.
   * @param a The first string to compare.  Cannot be null.
   * @param b The second string to compare.  Cannot be null.
   * @return true iff a and b are the same (ignoring case).
   * For example:<pre>
   *   equalsIgnoreCase( "hi", "HI" ) = true
   *   equalsIgnoreCase( "hi", "hi " ) = false
   *   equalsIgnoreCase( "", "" ) = true
   *    </pre>
   * @see String#equalsIgnoreCase(String)
   */
  public static boolean equalsIgnoreCase( String a, String b ) {
    checkForNull( a, "equalsIgnoreCase", "String" );
    checkForNull( b, "equalsIgnoreCase", "String" );
    return a.equalsIgnoreCase(b);
    }
  
  /** Return the number of characters in a string.
   * A static version of String.length(), for use in CS1 before we introduce objects.
   * @param _this The string to find the length of.  Cannot be null.
   * @return The number of characters in `_this`.
   * For example:<pre>
   *   length( "hi ho" ) = 5
   *   length( "z" ) = 1
   *   length( "" ) = 0
   *    </pre>
   * @see String#length()
   */
  public static int length( String _this ) {
    checkForNull( _this, "length", "String" );
    return _this.length();
    }
  /** Return a substring of the given string.
   * A static version of `String.substring(int,int)`, for use in CS1 before we introduce objects.
   * @param _this The `String` to take a substring from.  Cannot be null.
   * @param start The index of the first character of the substring.
   * @param stop The index of the first character after the substring.
   * @return A String consisting from characters at indices [`start`,`stop`) from `_this`.
   * @see String#substring(int,int)
   */
  public static String  substring( String _this, int from, int to ) {
    checkForNull( _this, "substring", "String" );
    return _this.substring(from,to);
    }
    
  /** Return a substring of the given string.
   * A static version of `String.substring(int)`, for use in CS1 before we introduce objects.
   * @param _this The `String` to take a substring from.  Cannot be null.
   * @param start The index of the first character of the substring.
   * @return A String consisting from characters at indices [`start`,`length(_this)`) from `_this`.
   * @see String#substring(int)
   */
  public static String  substring( String _this, int from ) { 
    checkForNull( _this, "substring", "String" );
    return _this.substring(from);
    }
    
    
  /** Return a substring of the given string.
   * A static version of `String.indexOf(String)`, for use in CS1 before we introduce objects.
   * @param _this The `String` to look inside.  Cannot be null.
   * @param target The String to try to find occuring inside `_this`.  Cannot be null.
   * @return The index of `_this` where `target` starts, or -1.
   *   That is: if int i=indexOf(s1,s2), then i==-1 || equals(s2,substring(s1,i,i+length(s2)))
   *   (equivalently, i==-1 || s2.equals(s1.substring(i,i+s2.length())))
   * 
   * @see String#indexOf(String)
   */
  public static int indexOf( String _this, String target ) {
    checkForNull( _this,  "indexOf", "String" );
    checkForNull( target, "indexOf", "String" );
    return _this.indexOf(target);
    }
    
  /** Return a lower-case version of the given string.
   * A static version of `String.toLowerCase()`, for use in CS1 before we introduce objects.
   * @param _this The `String` to take a substring from.  Cannot be null.
   * @return A lower-case version of `_this`.
   * @see String#toLowerCase()
   */
  public static String toLowerCase( String _this ) {
    checkForNull( _this, "toLowerCase", "String" );
    return _this.toLowerCase();
    }
  /** Return an upper-case version of the given string.
   * A static version of `String.toUpperCase()`, for use in CS1 before we introduce objects.
   * @param _this The `String` to convert to upper case.  Cannot be null.
   * @return An upper-case version of `_this`.
   * @see String#toUpperCase()
   */
  public static String toUpperCase( String _this ) {
    checkForNull( _this, "toUpperCase", "String" );
    return _this.toUpperCase();
    }
  /** Convert a String to an int.
   * @param s A string which is a valid representation of an int.
   * @return the int represented by `s`.
   * For example:<pre>
   *    stringToDouble("2") = 2
   *    stringToDouble("007") = 7
   *    </pre>
   * @throws NumberFormatException if `s` does not represent a valid double.
   *    stringToDouble("2+3") throws NumberFormatException
   */
  public static int stringToInt( String s ) {
    checkForNull( s, "stringToInt", "String" );
    return Integer.parseInt(s);
    }
    
  /** Convert a String to a double.
   * @param s A string which is a valid representation of a double.
   * @return the double represented by `s`.
   * For example:<pre>
   *    stringToDouble("43.2") = 43.2
   *    stringToDouble("2") = 2.0
   *    stringToDouble("007") = 7.0
   *    </pre>
   * @throws NumberFormatException if `s` does not represent a valid double.
   *    For example: <pre>
   *    stringToDouble("2+3") -> NumberFormatException
   *    </pre>
   */
  public static double  stringToDouble( String s ) {
    checkForNull( s, "stringToDouble", "String" );
    return Double.parseDouble(s);
    }
    
  /** Convert an int to a String.
   * @param n The int to convert to a String.
   * @return a String representing `n`.
   * For example:<pre>
   *    intToString(0) = "0"
   *    intToString(-2) = "-2"
   *    intToString(34) = "34"
   *    </pre>
   * @see Integer#toString()
   * @see #toString(Object)
   */
  public static String  intToString( int n ) { return (new Integer(n)).toString(); } // Redundant, with toString(Object).

  /** Convert an int to a double.
   *  (Same as casting to an int, but use the regular syntax for calling-a-method.)
   * @param n the int to convert.
   * @return n as a double.
   * For example:<pre>
   *    intToDouble(0) = 0.0
   *    intToDouble(-2) = -2.0
   *    intToDouble(2000000000) = 2e9
   *    intToDouble(2000000001) = 2.000000001e9
   *    </pre>
   */
  public static double  intToDouble( int n ) { return (double)n; }
  
  /** Convert a double to an int (truncating towards zero).
   *  (Same as casting to an int, but use the regular syntax for calling-a-method.)
   * @param x The double to convert to an int.
   * @return The int nearest to x, but between 0 and x.
   * For example:<pre>
   *    doubleToInt(0.0) = 0
   *    doubleToInt(3.1) = 3
   *    doubleToInt(3.9) = 3
   *    doubleToInt(-3.1) = -3.0
   *    doubleToInt(-3.9) = -3.0
   *    </pre>
   * @see Math#floor(double)
   * @see Math#ceil(double)
   * @see Math#round(double)
   */
  public static int doubleToInt( double x ) { return (int)Math.round(x); }

  /** Convert a String-of-length-1 to a char.
   * @param s A string of length 1.
   * @return s as a char.
   * @see String#charAt(int)
   */
  public static char    stringToChar( String s ) {
    checkForNull( s, "stringToChar", "String" );
    if (s.length() != 1) throw new IllegalArgumentException( "Object120.stringToChar: Can only convert strings of length 1; got: \"" + s + "\"." );
    return s.charAt(0);
    }

  /** Convert a char to a String-of-length-1.
   * @param c The char to convert.
   * @return c as a String.
   * @see Character#toString()
   * @see #toString(Object)
   */
  public static String charToString( char c ) {
    return Character.toString(c);
    }
    
  /** Select a single character from a string, by index (starting at 0).
   * A static version of `String.charAt(int)`, for use in CS1 before we introduce objects.
   * @param _this The `String` to select a character from.  Cannot be null.
   * @param i The index of the character to select from s; 0 <= i < length(s).
   * @return The `i`th character of `_this`.
   * @see String#toLowerCase()
   */
  public static char charAt( String _this, int i ) {
    checkForNull( _this, "charAt", "String" );
    return _this.charAt(i);
    }

  /** Return whether or not one String contains another.
   * A static version of `String.contains(CharSequence)`, for use in CS1 before we introduce objects.
   * @param _this The `String` to look inside of.  Cannot be null.
   * @param target The String to look for in `_this`.
   * @return whether or not `target` occurs in `_this`.
   * @see String#contains(CharSequence)
   * @see indexOf(String,String)
   * @see String#index(String)
   */
  public static boolean contains( String _this, CharSequence target ) {
    checkForNull( _this, "charAt", "String" );
    checkForNull( target, "charAt", "String (or CharSequence)" );
    return _this.contains(target);
    }
  /** Return whether or not a String is the empty string "" (0 letters long).
   * A static version of `String.isEmpty()`, for use in CS1 before we introduce objects.
   * @param _this The String to compare to "".  Cannot be null.
   * @return whether or not equals(_this,"") (equivalently, length(_this)==0).
   * @see String#isEmpty()
   */
  public static boolean isEmpty( String _this ) {
    checkForNull( _this, "isEmpty", "String" );
    return "".equals(_this);
    }

    
  /* Some common java.util.List methods. */
    
  /** Return whether or not a List is the empty list (0 items long).
   * A static version of `List.isEmpty()`, for use in CS1 before we introduce objects.
   * @param _this The List to check.  Cannot be null.
   * @return whether or not size(_this)==0.
   * @see List#isEmpty()
   */
  public static boolean isEmpty( List _this ) {
    checkForNull( _this, "isEmpty", "List" );
    return _this.isEmpty();
    }
  /** Return the size of a List -- the number of elements it contains
   * (including any duplicates).
   * A static version of `List.size()`, for use in CS1 before we introduce objects.
   * @param _this The List to check.  Cannot be null.
   * @return the number of items in `_this`.
   * @see List#size()
   */
  public static int size( List _this ) {
    checkForNull( _this, "size", "List" );
    return _this.size();
    }

  /** Add an element to a list.
   * A static version of `List.add(Object)`, for use in CS1 before we introduce objects.
   * @param _this The List to add an item to.  Cannot be null.
   * @param obj The item to add to `_this`.  Changes the contents of `_this`.
   * @return true.
   * @see List#add(Object)
   */
  public static <T,S extends T> boolean add( List<T> _this, S item ) {
    checkForNull( _this, "add", "List" );
    checkForNull( item, "add", "Object" );
    return _this.add(item);
    }

  /** Add an element to a list at a particular index,
   * shifting all items already at/after that index.
   * A static version of `List.add(int,Object)`, for use in CS1 before we introduce objects.
   * (The list is modified.)
   * @param _this The List to add an item to.  Cannot be null.
   * @param index The index to add the item at; must be in [0,size(_this)].
   *   (If index == size(_this), then item is added to the end of _this.)
   * @param obj The item to add to `_this`.
   * @see List#add(Object)
   */
  public static <T,S extends T> void add( List<T> _this, int index, S item ) {
    checkForNull( _this, "add", "List" );
    checkForNull( item, "add", "Object" );
    _this.add(index,item);
    }
    
  /** Add an element to a list at a particular index,
   * shifting all items already at/after that index.
   * A static version of `List.add(int,Object)`, for use in CS1 before we introduce objects.
   * (The list is modified.)
   * @param _this The List to add an item to.  Cannot be null.
   * @param index The index to add the item at; must be in [0,size(_this)].
   *   (If index == size(_this), then item is added to the end of _this.)
   * @param obj The item to add to `_this`.
   * @return true.
   * @see List#add(Object)
   */
  public static <T,S extends T> boolean addAll( List<T> _this, Collection<S> items ) {
    checkForNull( _this, "addAll", "List" );
    checkForNull( items, "addAll", "Collection" );
    return _this.addAll(items);
    }
    
    
  /* Deficiencies of this class:
   * Fundamental: when extending Object120:
   *  - final fields won't compile -- not init'd.
   *  - Can't generate setters (only a `set(String fieldName, Object val)`; no good for CS1.)
   * To fix/patch:
   * 
   * Really, this class should be broken into several others:
   *   Util120 for utility functions, and trio of classes Object,MutableObject,ImmutableObject.
   * However, for CS1 I only wanted just a single file that Students need to include, not two-four.
   */
  }
