import greenfoot.*;  // (World, Actor, GreenfootImage, and Greenfoot)
import java.awt.Color;

/** A basic turtle.
 *
 * @author Poul Henriksen, procedural(static) version by Ian Barland.
 * @version 1.0.2
 */
public class Trtl extends Actor {

      
    /** Moves the Trtl the given distance.
     * @param _this The Trtl to be moved.
     * @param distance The number of pixels to move the turtle
     *                 in the direction it's facing.
     */
    public static void move(Trtl _this, double distance) {
        double directionRad = Math.toRadians(_this.direction);
        double xDist = distance * Math.cos(directionRad);
        double yDist = distance * Math.sin(directionRad);
        moveTo( _this, _this.x + xDist,  _this.y + yDist );
    }

    /** Takes the pen up.
     * @param _this The Trtl whose pen is to be lifted.
     */
    public static void penUp(Trtl _this) {
        _this.penDown = false;
        _this.setImage(_this.penUpImage);
    }

    /** Puts the pen down.
     * @param _this The Trtl whose pen is to be lowered.
     */
    public static void penDown(Trtl _this) {
        _this.penDown = true;
        _this.setImage(_this.penDownImage);
    }
    
    /** Turns a turtle.
     * @param _this The Trtl to be turned.
     * @param degrees How many degrees clockwise to turn the turtle.
     */
    public static void turn( Trtl _this, double degrees) {
        _this.direction = _this.direction + degrees;
        if (! (0 <= _this.direction && _this.direction < MAX_ANGLE)) {
            _this.direction = _this.direction % MAX_ANGLE;
        }
        _this.setRotation((int) _this.direction);
    }
    
    /** Moves the Trtl to the given position,
     * drawing iff pen is down.
     */
    private static void moveTo( Trtl _this, double newX, double newY) {
        if (_this.penDown) {
            drawLine(_this, _this.x, _this.y, newX, newY);
        }
        _this.x = newX;
        _this.y = newY;
        setLocation(_this, _this.x, _this.y);
    }


    /** Set the color of the pen.
     * @param _this The Trtl whose pen-color is being set.
     * @param newColor The name of the color to use;
     *  accepted colours are: 
     *  "red", "black", "blue", "yellow", "green", "magenta", "white"
     */
    public static void setColor(Trtl _this, String newColor) {
        _this.color = newColor;
        addDot(_this.penDownImage, newColor);
    }
    
    /** Translate a String into a Color
     */
    private static Color decode(String colorString) {
        if(colorString.equals("red"))
            return Color.red;
        else if(colorString.equals("black"))
            return  Color.black;
        else if(colorString.equals("blue"))
            return Color.blue;
        else if(colorString.equals("yellow"))
            return  Color.yellow;
        else if(colorString.equals("green"))
            return Color.green;
        else if(colorString.equals("magenta"))
            return Color.magenta;
        else if(colorString.equals("white"))
            return Color.white;
        else
            return Color.black;
    }
    
    
    /** A hook called when this Trtl is added to a World.
     */
    protected void addedToWorld(World world) {
        /* We need to make sure that our own notion of the location
         * is the same as the World's.
         * (Not procedural, since it's called by the Greenfoot environment.)
         */
        this.x = getX();
        this.y = getY();
    }

    /** Draw a line between two points with the Trtl's current colour.
     */
    private static void drawLine(Trtl _this, double x1, double y1, double x2, double y2) {
        GreenfootImage image = _this.getWorld().getBackground();
        Color awtColor = decode(_this.color);
        image.setColor(awtColor);
        image.drawLine((int) Math.ceil(x1) , (int) Math.ceil(y1), (int) Math.ceil(x2), (int) Math.ceil(y2));
    }

    /** Modify `image` to include a `color`ed dot.
     */
    private static void addDot(GreenfootImage image, String color) {
        double halfWidth = image.getWidth() / 2.;
        double halfHeight = image.getHeight() / 2.;
        int penWidth = (int) halfWidth / 2;
        int penHeight = (int) halfHeight / 2;
        int penX = (int) (halfWidth - penWidth / 2);
        int penY = (int) (halfHeight - penHeight / 2);
        Color awtColor = decode(color);
        image.setColor(awtColor);
        image.fillOval(penX - 3, penY - 1, penWidth, penHeight);
    }
    
    /** Set the location of the Trtl _this.
     *  @see setLocation(int,int)
     */
    protected static void setLocation(Trtl _this, double x, double y) {
        _this.x = x;
        _this.y = y;
        _this.setLocation((int) Math.floor(x), (int) Math.floor(y));
    }
    
    /** Set the location of the turtle.
     * We must override setLocation(int,int) so we can interactively move objects.
     * Not procedural, since it's called by the Greenfoot environment.
     */
    public void setLocation(int x, int y) {
        // Update fields, unless it would only introduce rounding error.
        if (x != (int)Math.floor(this.x)) {
            this.x = x;
        }
        if (y != (int)Math.floor(this.y)) {
            this.y = y;
        }
        super.setLocation( x, y );
    }
    
    /* Construct a basic turtle.  Non-public. */
    protected Trtl() {
        penUpImage = this.getImage();
        penDownImage = new GreenfootImage(this.penUpImage);
        addDot(this.penDownImage, this.color);
        penDown(this);    // Caution: calling method from constructor.
    }

    /** Create a new Trtl.
     * @return a new Trtl.
     */
    public static Trtl makeTrtl() {
        return addToWorld( new Trtl() );
    }

    public static <T extends Actor> T addToWorld( T a ) {
        TurtleWorld wrld = TurtleWorld.theWorld;
        int w = a.getImage().getWidth();
        int h = a.getImage().getHeight();
        int SPREAD = 3;  // how widely scattered new objects are.
        int dw = Greenfoot.getRandomNumber(w*SPREAD) - w*SPREAD/2;
        int dh = Greenfoot.getRandomNumber(h*SPREAD) - h*SPREAD/2;
        
        wrld.addObject(a,
                       wrld.getWidth()/5  + dw,
                       wrld.getHeight()/2 + dh );
        return a;
    }
    
    /** Whether the turtle should paint or not. */
    private boolean penDown;
    
    /** Colour of the pen. */
    private String color = "black";
    
    /** The direction the turtle is facing. */
    public double direction;
    
    /** The x location.  Will be initialized via addedToWorld. */
    private double x;
    
    /** The y location.  Will be initialized via addedToWorld. */
    private double y;
    
    /** The maximum possible positive angle. */
    private static double MAX_ANGLE = 360;
     
    /** Image used when the pen is up */
    private GreenfootImage penUpImage;
    
    /** Image used when the pen is down */
    private GreenfootImage penDownImage;
    
}